##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from io import StringIO
from tempfile import NamedTemporaryFile
from unittest import TestCase

from ruamel.yaml import YAML

from exhibition.main import Config


YAML_DATA = """
sitename: bob
thingy:
    - one
    - two
    - three
"""


class ConfigTestCase(TestCase):
    def test_load_str(self):
        settings = Config(YAML_DATA)

        self.assertEqual(len(settings), 2)
        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_load_file(self):
        yaml_file = StringIO(YAML_DATA)
        settings = Config(yaml_file)

        self.assertEqual(len(settings), 2)
        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_load_dict(self):
        yaml_dict = dict(YAML(typ="safe").load(YAML_DATA))
        settings = Config(yaml_dict)

        self.assertEqual(len(settings), 2)
        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_no_load(self):
        settings = Config()
        self.assertEqual(len(settings), 0)

    def test_from_path(self):
        with NamedTemporaryFile() as yaml_file:
            yaml_file.write(YAML_DATA.encode())
            yaml_file.file.flush()  # just to be sure

            settings = Config.from_path(yaml_file.name)

            self.assertEqual(len(settings), 2)
            self.assertEqual(settings["sitename"], "bob")
            self.assertEqual(settings["thingy"], ["one", "two", "three"])

    def test_getitem(self):
        settings = Config(YAML_DATA)

        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])

        self.assertEqual(settings.get("sitename"), "bob")
        self.assertEqual(settings.get("sitetitle"), None)

        with self.assertRaises(KeyError):
            settings["sitetitle"]

    def test_getitem_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent)

        self.assertEqual(settings["sitename"], "bob")
        self.assertEqual(settings["thingy"], ["one", "two", "three"])
        self.assertEqual(settings["test"], True)

        self.assertEqual(settings.get("sitename"), "bob")
        self.assertEqual(settings.get("test"), True)
        self.assertEqual(settings.get("sitetitle"), None)
        self.assertEqual(parent.get("sitename"), None)

        with self.assertRaises(KeyError):
            settings["sitetitle"]

    def test_keys(self):
        settings = Config(YAML_DATA)

        self.assertCountEqual(list(settings.keys()), ["sitename", "thingy"])

    def test_keys_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent)

        self.assertCountEqual(list(settings.keys()), ["sitename", "thingy", "test"])
        self.assertCountEqual(list(parent.keys()), ["test"])

    def test_values(self):
        settings = Config(YAML_DATA)

        self.assertCountEqual(list(settings.values()), ["bob", ["one", "two", "three"]])

    def test_values_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent)

        self.assertCountEqual(list(settings.values()), ["bob", ["one", "two", "three"], True])
        self.assertCountEqual(list(parent.values()), [True])

    def test_items(self):
        settings = Config(YAML_DATA)

        self.assertCountEqual(list(settings.items()),
                              [("sitename", "bob"), ("thingy", ["one", "two", "three"])])

    def test_items_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent)

        self.assertCountEqual(list(settings.items()),
                              [("sitename", "bob"), ("thingy", ["one", "two", "three"]),
                               ("test", True)])
        self.assertCountEqual(list(parent.items()), [("test", True)])

    def test_contains(self):
        settings = Config(YAML_DATA)

        self.assertIn("sitename", settings)
        self.assertIn("thingy", settings)
        self.assertNotIn("test", settings)

    def test_contains_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent)

        self.assertIn("test", settings)
        self.assertIn("sitename", settings)
        self.assertIn("thingy", settings)

        self.assertIn("test", parent)
        self.assertNotIn("sitename", parent)
        self.assertNotIn("thingy", parent)

    def test_len(self):
        settings = Config(YAML_DATA)

        self.assertEqual(len(settings), 2)

    def test_len_with_parent(self):
        parent = Config({"test": True})
        settings = Config(YAML_DATA, parent=parent)

        self.assertEqual(len(settings), 3)
        self.assertEqual(len(parent), 1)



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from unittest import TestCase, mock
import logging

from click.testing import CliRunner

from exhibition import command, main


class CommandTestCase(TestCase):
    @mock.patch("exhibition.command.main.gen")
    @mock.patch("exhibition.command.main.Config.from_path", return_value=main.Config())
    def test_gen(self, config_mock, gen_mock):
        runner = CliRunner()
        result = runner.invoke(command.exhibition, ["gen"])

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(gen_mock.call_count, 1)
        self.assertEqual(gen_mock.call_args, ((config_mock.return_value,), {}))

        self.assertEqual(config_mock.call_args, ((main.SITE_YAML_PATH,), {}))

    @mock.patch("exhibition.command.main.serve", return_value=(mock.Mock(), mock.Mock()))
    @mock.patch("exhibition.command.main.Config.from_path", return_value=main.Config())
    def test_serve(self, config_mock, serve_mock):
        runner = CliRunner()
        result = runner.invoke(command.exhibition, ["serve"])

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(serve_mock.call_count, 1)
        self.assertEqual(serve_mock.call_args, ((config_mock.return_value,), {}))

        self.assertEqual(config_mock.call_args, ((main.SITE_YAML_PATH,), {}))
        self.assertEqual(serve_mock.return_value[0].shutdown.call_count, 0)
        self.assertEqual(serve_mock.return_value[1].join.call_count, 1)

        serve_mock.return_value[1].join.side_effect = KeyboardInterrupt
        # this time with a simulated keyboard interrupt
        result = runner.invoke(command.exhibition, ["serve"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(serve_mock.call_count, 2)
        self.assertEqual(serve_mock.call_args, ((config_mock.return_value,), {}))

        self.assertEqual(config_mock.call_args, ((main.SITE_YAML_PATH,), {}))
        self.assertEqual(serve_mock.return_value[0].shutdown.call_count, 1)
        self.assertEqual(serve_mock.return_value[1].join.call_count, 2)

    @mock.patch.object(command, "logger")
    def test_exhibition(self, log_mock):
        @command.exhibition.command()
        def test():
            pass

        # logging defaults to warn
        runner = CliRunner()
        result = runner.invoke(command.exhibition, ["test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.WARN,), {}))

        # one verbose is info
        log_mock.reset_mock()
        result = runner.invoke(command.exhibition, ["-v", "test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.INFO,), {}))

        # two or more verbose is debug
        log_mock.reset_mock()
        result = runner.invoke(command.exhibition, ["-vv", "test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_count, 1)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.DEBUG,), {}))

        result = runner.invoke(command.exhibition, ["-vvv", "test"])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(log_mock.addHandler.call_count, 2)
        self.assertEqual(log_mock.setLevel.call_count, 2)
        self.assertEqual(log_mock.setLevel.call_args, ((logging.DEBUG,), {}))



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from tempfile import TemporaryDirectory
from unittest import TestCase
import pathlib

from exhibition.main import Node


GOOD_META = """---
thingy: 3
---
Some text
"""

BAD_END_META = """---
thingy: 3

Some text
"""

BAD_START_META = """
---
thingy: 3
---
Some text
"""

NO_META = """
Some text
"""

YAML_FILE = """
thingy: 3
bob:
    - 1
    - 2
"""

JSON_FILE = """
{
    "thingy": 3,
    "bob": [
        1,
        2
    ]
}
"""

BINARY_FILE = (
    b"\x00\x00\x01\x00\x02\x00\x10\x10\x10\x00\x01\x00\x04\x00(\x01\x00\x00&"
    b"\x00\x00\x00  \x10\x00\x01\x00\x04\x00\xe8\x02\x00\x00N\x01\x00\x00("
    b"\x00\x00\x00\x10\x00\x00\x00 \x00\x00\x00\x01\x00\x04\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x01\x00\x00\x08\x0c\n\x00\x16\x18\x17\x00:<;\x00"
    b"@BA\x00TV"
)


class NodeTestCase(TestCase):
    def setUp(self):
        self.content_path = TemporaryDirectory()
        self.deploy_path = TemporaryDirectory()

        self.default_settings = {"content_path": self.content_path.name,
                                 "deploy_path": self.deploy_path.name}

    def tearDown(self):
        self.content_path.cleanup()
        self.deploy_path.cleanup()

    def test_repr(self):
        path = pathlib.Path(self.content_path.name, "page.html")
        path.touch()
        node = Node(path, None, meta=self.default_settings)

        self.assertEqual(repr(node), "<Node: page.html>")

    def test_full_path(self):
        path = pathlib.Path(self.content_path.name, "page.html")
        path.touch()
        node = Node(path, None, meta=self.default_settings)

        self.assertEqual(node.full_path, self.deploy_path.name)

    def test_full_path_with_parent(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_path, self.deploy_path.name + "/page.html")

    def test_full_url(self):
        path = pathlib.Path(self.content_path.name, "page.html")
        path.touch()
        node = Node(path, None, meta=self.default_settings)

        self.assertEqual(node.full_url, "/")

    def test_full_url_with_parent(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.jpg")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_url, "/page.jpg")

    def test_full_url_with_strip_ext(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_url, "/page")

    def test_full_url_with_index(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "index.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(child_node.full_url, "/")

    def test_walk(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "index.html")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(list(parent_node.walk(include_self=True)), [parent_node, child_node])
        self.assertEqual(list(parent_node.walk()), [child_node])

        self.assertEqual(list(child_node.walk(include_self=True)), [child_node])
        self.assertEqual(list(child_node.walk()), [])

    def test_render_dir(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog")
        child_path.mkdir()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertFalse(pathlib.Path(child_node.full_path).exists())
        child_node.render()
        self.assertTrue(pathlib.Path(child_node.full_path).exists())
        self.assertTrue(pathlib.Path(child_node.full_path).is_dir())

    def test_render_file(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog")
        child_path.touch()

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertFalse(pathlib.Path(child_node.full_path).exists())
        child_node.render()
        self.assertTrue(pathlib.Path(child_node.full_path).exists())
        self.assertTrue(pathlib.Path(child_node.full_path).is_file())

    def test_render_binary_file(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "blog")
        with child_path.open("wb") as cf:
            cf.write(BINARY_FILE)

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertFalse(pathlib.Path(child_node.full_path).exists())
        child_node.render()
        self.assertTrue(pathlib.Path(child_node.full_path).exists())
        self.assertTrue(pathlib.Path(child_node.full_path).is_file())

        with pathlib.Path(child_node.full_path).open("rb") as df:
            content = df.read()
            self.assertEqual(content, BINARY_FILE)

    def test_process_good_meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(GOOD_META)

        node = Node(path, None)
        self.assertEqual(list(node._meta.keys()), [])
        self.assertEqual(node._content_start, None)

        self.assertEqual(list(node.meta.keys()), ["thingy"])
        self.assertEqual(node._content_start, 18)

    def test_process_bad_start_meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(BAD_START_META)

        node = Node(path, None)
        self.assertEqual(list(node._meta.keys()), [])
        self.assertEqual(node._content_start, None)

        self.assertEqual(list(node.meta.keys()), [])
        self.assertEqual(node._content_start, 0)

    def test_process_bad_end__meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(BAD_END_META)

        node = Node(path, None)
        self.assertEqual(list(node._meta.keys()), [])
        self.assertEqual(node._content_start, None)

        self.assertEqual(list(node.meta.keys()), [])
        self.assertEqual(node._content_start, 0)

    def test_process_no_meta(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write(NO_META)

        node = Node(path, None)
        self.assertEqual(list(node._meta.keys()), [])
        self.assertEqual(node._content_start, None)

        self.assertEqual(list(node.meta.keys()), [])
        self.assertEqual(node._content_start, 0)

    def test_get_content(self):
        path = pathlib.Path(self.content_path.name, "blog")
        with path.open("w") as f:
            f.write("a" * 10)

        node = Node(path, None)
        self.assertEqual(node.get_content(), "a" * 10)
        self.assertEqual(node._content_start, 0)

        node._content = None
        node._content_start = 5
        node.get_content()
        self.assertEqual(node.get_content(), "a" * 5)
        self.assertEqual(node._content_start, 5)

    def test_yaml_data(self):
        path = pathlib.Path(self.content_path.name, "blog.yaml")
        with path.open("w") as f:
            f.write(YAML_FILE)

        node = Node(path, None)

        self.assertEqual(node.data, {"thingy": 3, "bob": [1, 2]})

    def test_json_data(self):
        path = pathlib.Path(self.content_path.name, "blog.json")
        with path.open("w") as f:
            f.write(JSON_FILE)

        node = Node(path, None)

        self.assertEqual(node.data, {"thingy": 3, "bob": [1, 2]})

    def test_dir_data(self):
        path = pathlib.Path(self.content_path.name, "blog.json")
        path.mkdir()

        node = Node(path, None)

        self.assertEqual(node.data, None)

    def test_data_unsupported(self):
        path = pathlib.Path(self.content_path.name, "blog.xml")
        path.touch()

        node = Node(path, None)

        with self.assertRaises(KeyError):
            node.data

    def test_add_child(self):
        parent_path = pathlib.Path(self.content_path.name)
        child_path = pathlib.Path(self.content_path.name, "page.html")

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child_node = Node(child_path, parent_node)

        self.assertEqual(parent_node.children, {"page.html": child_node})

        with self.assertRaises(AssertionError):
            parent_node.add_child(Node(pathlib.Path(self.content_path.name, "page2.html"), None))

    def test_siblings(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child2_path = pathlib.Path(self.content_path.name, "page2.html")

        parent_node = Node(parent_path, None, meta=self.default_settings)
        child1_node = Node(child1_path, parent_node)
        child2_node = Node(child2_path, parent_node)

        self.assertEqual(child1_node.siblings, {"page2.html": child2_node})
        self.assertEqual(child2_node.siblings, {"page1.html": child1_node})

    def test_from_path(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page2.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path)
        self.assertCountEqual(list(parent_node.children.keys()), ["page1.html", "page2.html"])

    def test_from_path_and_ignore(self):
        parent_path = pathlib.Path(self.content_path.name)
        child1_path = pathlib.Path(self.content_path.name, "page1.html")
        child1_path.touch()
        child2_path = pathlib.Path(self.content_path.name, "page2.html")
        child2_path.touch()

        parent_node = Node.from_path(parent_path, meta={"ignore": "*.html"})
        self.assertCountEqual(list(parent_node.children.keys()), [])



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

"""
Documentation for this module can be found in :doc:`commandline`
"""

import logging

import click

from exhibition import main


logger = logging.getLogger("exhibition")


@click.group()
@click.option("-v", "--verbose", count=True,
              help="Verbose output, can be used multiple times to increase logging level")
def exhibition(verbose):
    logger.addHandler(logging.StreamHandler())
    if verbose > 1:
        logger.setLevel(logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARN)


@exhibition.command(short_help="Generate site")
def gen():
    """
    Generate site from content_path
    """
    settings = main.Config.from_path(main.SITE_YAML_PATH)
    main.gen(settings)


@exhibition.command(short_help="Serve site locally")
def serve():
    """
    Serve files from deploy_path as a webserver would
    """
    settings = main.Config.from_path(main.SITE_YAML_PATH)
    httpd, thread = main.serve(settings)

    try:
        thread.join()
    except (KeyboardInterrupt, SystemExit):
        httpd.shutdown()



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from tempfile import TemporaryDirectory
from unittest import TestCase
import os
import pathlib

from exhibition.main import gen, Config


class GenTestCase(TestCase):
    def test_clean_deploy_dir(self):
        with TemporaryDirectory() as deploy, TemporaryDirectory() as content:
            settings = Config({"deploy_path": deploy, "content_path": content})
            old_file = os.path.join(deploy, "someoldfile")
            with open(old_file, "w") as of:
                of.write("content!")

            # file exists and then is deleted during site generation
            self.assertTrue(os.path.exists(old_file))
            gen(settings)
            self.assertFalse(os.path.exists(old_file))

    def test_walk_nodes(self):
        files = [
            "blog/index.html",
            "blog/post.html",
            "index.html",
            "style.css",
        ]
        dirs = ["blog"]
        with TemporaryDirectory() as deploy, TemporaryDirectory() as content:
            settings = Config({"deploy_path": deploy, "content_path": content})
            for d in dirs:
                pathlib.Path(content, d).mkdir()

            for f in files:
                pathlib.Path(content, f).touch()

            gen(settings)

            for item in files + dirs:
                with self.subTest("%s exists in %s" % (item, deploy)):
                    self.assertTrue(pathlib.Path(deploy, item).exists())



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from http.client import HTTPConnection
from tempfile import TemporaryDirectory
from unittest import TestCase
import os

from exhibition.main import Config, serve


INDEX_CONTENTS = """<html>
    <head>
        <title>Hello</title>
    </head>
    <body>
        Hello world!
    </body>
</html>
"""

CSS_CONTENTS = """// CSS file
html, body {
    color: black;
    background-color: white;
}
"""


class ServeTestCase(TestCase):
    def setUp(self):
        self.tmp_dir = TemporaryDirectory()

        with open(os.path.join(self.tmp_dir.name, "index.html"), "w") as index:
            index.write(INDEX_CONTENTS)

        with open(os.path.join(self.tmp_dir.name, "style.css"), "w") as css:
            css.write(CSS_CONTENTS)

        self.client = HTTPConnection("localhost", "8000")

    def tearDown(self):
        self.tmp_dir.cleanup()
        self.client.close()
        self.server.shutdown()
        self.server.server_close()

    def get_server(self, settings):
        httpd, thread = serve(settings)
        self.server = httpd

    def test_fetch_file(self):
        settings = Config({"deploy_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/style.css")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, CSS_CONTENTS.encode())

    def test_fetch_file_with_prefix(self):
        settings = Config({"deploy_path": self.tmp_dir.name, "base_url": "/bob/"})
        self.get_server(settings)

        self.client.request("GET", "/bob/style.css")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, CSS_CONTENTS.encode())

        self.client.request("GET", "/style.css")
        response = self.client.getresponse()
        self.assertEqual(response.status, 404)

    def test_index(self):
        settings = Config({"deploy_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/")
        response = self.client.getresponse()
        self.assertEqual(response.status, 200)
        content = response.read()
        self.assertEqual(content, INDEX_CONTENTS.encode())

    def test_404(self):
        settings = Config({"deploy_path": self.tmp_dir.name})
        self.get_server(settings)

        self.client.request("GET", "/not-existing.html")
        response = self.client.getresponse()
        self.assertEqual(response.status, 404)



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from tempfile import TemporaryDirectory
from unittest import TestCase, mock
import pathlib

from jinja2.exceptions import TemplateRuntimeError
from jinja2 import Markup

from exhibition.filters.jinja2 import content_filter as jinja_filter
from exhibition.main import Node


PLAIN_TEMPLATE = """
{% for i in range(3) -%}
    <p>{{ i }}</p>
{%- endfor %}
""".strip()

CONTENT_BLOCK = "".join(["{% block content %}", PLAIN_TEMPLATE, "{% endblock %}"])

BASE_TEMPLATE = """
<p>Title</p>
{% block content %}{% endblock %}
""".strip()

ERROR_TEMPLATE = """
{% raise "This is an error" %}
""".strip()

MARK_TEMPLATE = """
{% mark thingy %}
Hello
{% endmark %}
Bye
""".strip()

MD_TEMPLATE = """
{% filter markdown %}
## Hello

This is *text*
{% endfilter %}
""".strip()

TYPOG_TEMPLATE = """
{% filter typogrify %}
Hello -- how are you?
{% endfilter %}
""".strip()

METASORT_TEMPLATE = """
{% for child in node.children.values()|metasort("bob") -%}
{{ child.meta.bob }}
{%- endfor %}
""".strip()

METASELECT_TEMPLATE = """
{% for child in node.children.values()|metaselect("bob") -%}
{{ child.meta.bob }}
{%- endfor %}
""".strip()

METAREJECT_TEMPLATE = """
{% for child in node.children.values()|metareject("bob") -%}
{{ child.meta.bob }}
{%- endfor %}
""".strip()


class Jinja2TestCase(TestCase):
    def test_template(self):
        node = Node(mock.Mock(), None, meta={"templates": []})
        node.is_leaf = False
        result = jinja_filter(node, PLAIN_TEMPLATE)
        self.assertEqual(result, "<p>0</p><p>1</p><p>2</p>")

    def test_extends(self):
        template_name = "bob.j2"
        with TemporaryDirectory() as tmp_dir:
            with pathlib.Path(tmp_dir, template_name).open("w") as tmpl:
                tmpl.write(BASE_TEMPLATE)

            node = Node(mock.Mock(), None, meta={"templates": [tmp_dir], "extends": "bob.j2"})
            node.is_leaf = False
            result = jinja_filter(node, CONTENT_BLOCK)
            self.assertEqual(result, "<p>Title</p>\n<p>0</p><p>1</p><p>2</p>")

    def test_extends_with_multiple_template_dirs(self):
        template_name = "bob.j2"
        with TemporaryDirectory() as tmp1_dir, TemporaryDirectory() as tmp2_dir:
            with pathlib.Path(tmp1_dir, template_name).open("w") as tmpl:
                tmpl.write(BASE_TEMPLATE)

            with pathlib.Path(tmp2_dir, template_name).open("w") as tmpl:
                tmpl.write("empty")

            node = Node(mock.Mock(), None,
                        meta={"templates": [tmp1_dir, tmp2_dir], "extends": "bob.j2"})
            node.is_leaf = False
            result = jinja_filter(node, CONTENT_BLOCK)
            self.assertEqual(result, "<p>Title</p>\n<p>0</p><p>1</p><p>2</p>")

            # switch order of template dirs,
            node = Node(mock.Mock(), None,
                        meta={"templates": [tmp2_dir, tmp1_dir], "extends": "bob.j2"})
            node.is_leaf = False
            result = jinja_filter(node, CONTENT_BLOCK)
            self.assertEqual(result, "empty")

    def test_default_block(self):
        template_name = "bob.j2"
        with TemporaryDirectory() as tmp_dir:
            with pathlib.Path(tmp_dir, template_name).open("w") as tmpl:
                tmpl.write(BASE_TEMPLATE)

            node = Node(mock.Mock(), None, meta={"templates": [tmp_dir], "extends": "bob.j2"})
            node.is_leaf = False
            result = jinja_filter(node, PLAIN_TEMPLATE)
            self.assertEqual(result, "<p>Title</p>\n")

            node = Node(mock.Mock(), None, meta={"templates": [tmp_dir],
                                                 "extends": "bob.j2", "default-block": "content"})
            node.is_leaf = False
            result = jinja_filter(node, PLAIN_TEMPLATE)
            self.assertEqual(result, "<p>Title</p>\n\n<p>0</p><p>1</p><p>2</p>")

    def test_context(self):
        path = mock.Mock()
        path.name = "thisfile.html"
        node = Node(path, None, meta={"templates": []})
        node.is_leaf = False
        result = jinja_filter(node, "{{ node }}")
        self.assertEqual(result, "&lt;Node: thisfile.html&gt;")

        result = jinja_filter(node, "{{ undef_value }}")
        self.assertEqual(result, "")

    def test_raise_extension(self):
        node = Node(mock.Mock(), None, meta={"templates": []})
        node.is_leaf = False
        with self.assertRaises(TemplateRuntimeError) as excp:
            jinja_filter(node, ERROR_TEMPLATE)

        self.assertEqual(excp.exception.message, "This is an error")

    def test_mark_extension(self):
        with TemporaryDirectory() as content_path, TemporaryDirectory() as deploy_path:
            path = pathlib.Path(content_path, "blog.html")
            with path.open("w") as f:
                f.write(MARK_TEMPLATE)

            node = Node(path, Node(path.parent, None, {"content_path": content_path,
                                                       "deploy_path": deploy_path,
                                                       "filter": "exhibition.filters.jinja2",
                                                       "templates": []}))
            self.assertEqual(node.marks, {"thingy": Markup("\nHello\n")})

            node.render()
            with pathlib.Path(deploy_path, "blog.html").open("r") as f:
                content = f.read()

            self.assertEqual(content, "\nHello\n\nBye")

    def test_markdown_filter(self):
        node = Node(mock.Mock(), None, meta={"templates": []})
        node.is_leaf = False
        result = jinja_filter(node, MD_TEMPLATE)
        self.assertEqual(result, "<h2>Hello</h2>\n<p>This is <em>text</em></p>")

    def test_typogrify_filter(self):
        node = Node(mock.Mock(), None, meta={"templates": []})
        node.is_leaf = False
        result = jinja_filter(node, TYPOG_TEMPLATE)
        self.assertEqual(result, "\nHello &#8212; how are&nbsp;you?\n")

    def test_metasort(self):
        node = Node(mock.Mock(), None, meta={"templates": []})
        node._content_start = 0
        for i in [3, 5, 27, 2, 1]:
            new_node = Node(mock.Mock(), node, meta={"bob": i})
            new_node._content_start = 0
        result = jinja_filter(node, METASORT_TEMPLATE)

        self.assertEqual(result, "123527")

    def test_metaselect(self):
        node = Node(mock.Mock(),  None, meta={"templates": []})
        node._content_start = 0
        for i in [True, True, False, True, None]:
            new_node = Node(mock.Mock(), node, meta={"bob": i})
            new_node._content_start = 0
        result = jinja_filter(node, METASELECT_TEMPLATE)

        self.assertEqual(result, "TrueTrueTrue")

    def test_metareject(self):
        node = Node(mock.Mock(),  None, meta={"templates": []})
        node._content_start = 0
        for i in [True, True, False, True, None]:
            new_node = Node(mock.Mock(), node, meta={"bob": i})
            new_node._content_start = 0
        result = jinja_filter(node, METAREJECT_TEMPLATE)

        # dict key order isn't stable, but it'll be one of these
        self.assertIn(result, ["FalseNone", "NoneFalse"])



##
#
# Copyright (C) 2018 Matt Molyneaux
#
# This file is part of Exhibition.
#
# Exhibition is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Exhibition is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Exhibition.  If not, see <https://www.gnu.org/licenses/>.
#
##

from http.server import HTTPServer, SimpleHTTPRequestHandler
from importlib import import_module
import io
import logging
import os
import pathlib
import shutil
import threading

from ruamel.yaml import YAML

logger = logging.getLogger("exhibition")

SITE_YAML_PATH = "site.yaml"

yaml_parser = YAML(typ="safe")

DATA_EXTRACTORS = {
    ".yaml": yaml_parser.load,
    ".json": yaml_parser.load,
}


class Config:
    """
    Configuration object that implements a dict-like interface

    If a key cannot be found in this instance, the parent :class:`Config` will
    be searched (and its parent, etc.)
    """
    def __init__(self, data=None, parent=None):
        """
        :param data:
            Can be one of a string, a file-like object, a dict-like object, or
            ``None``. The first two will be assumed as YAML
        :param parent:
            Parent :class:`Config` or ``None`` if this is the root configuration object
        """
        self.parent = parent
        self._base_config = {}

        if data:
            self.load(data)

    def load(self, data):
        """
        Load data into configutation object

        :param data:
            If a string or file-like object, ``data`` is parsed as if it were
            YAML data. If a dict-like object, ``data`` is added to the internal
            dictionary.

            Otherwise an :class:`AssertionError` exception is raised
        """
        if isinstance(data, (str, io.IOBase)):
            self._base_config.update(yaml_parser.load(data))
        elif isinstance(data, dict):
            self._base_config.update(data)
        else:
            raise AssertionError("data needs to be a string, file-like, or dict-like object")

    @classmethod
    def from_path(cls, path):
        """Load YAML data from a file"""
        with open(path) as f:
            obj = cls(f)

        return obj

    def __getitem__(self, key):
        try:
            return self._base_config[key]
        except KeyError:
            if self.parent is None:
                raise
            else:
                return self.parent[key]

    def __setitem__(self, key, value):
        self._base_config[key] = value

    def __contains__(self, key):
        return key in self.keys()

    def __len__(self):
        return len(list(self.keys()))

    def __iter__(self):
        return self.keys()

    def keys(self):
        _keys_set = set()
        for k in self._base_config.keys():
            _keys_set.add(k)
            yield k

        if self.parent is not None:
            for k in self.parent.keys():
                if k not in _keys_set:
                    _keys_set.add(k)
                    yield k

    def values(self):
        for k in self.keys():
            yield self[k]

    def items(self):
        for k in self.keys():
            yield (k, self[k])

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def update(self, *args, **kwargs):
        self._base_config.update(*args, **kwargs)

    def copy(self):
        klass = type(self)
        return klass(self._base_config.copy(), self.parent)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self._base_config.keys())


class Node:
    """
    A node represents a file or directory
    """
    _meta_names = ["meta.yaml", "meta.yml"]
    _index_file = "index.html"
    _strip_exts = [".html"]

    _meta_header = "---\n"
    _meta_footer = "---\n"

    _dir_mode = 0o755
    _file_mode = 0o644

    _content_start = None
    _content = None
    _data = None
    _marks = None

    def __init__(self, path, parent, meta=None):
        """
        :param path:
            A :class:`pathlib.Path` that is either the ``content_path`` or a child of it.
        :param parent:
            Either another :class:`Node` or ``None``
        :param meta:
            A dict-like object that will be passed to a :class:`Config` instance
        """
        self.path_obj = path
        self.parent = parent
        self.children = {}

        self.is_leaf = self.path_obj.is_file()

        self._meta = Config({}, getattr(self.parent, "meta", None))
        if meta:
            self._meta.update(meta)

        if self.parent:
            self.parent.add_child(self)

    @classmethod
    def from_path(cls, path, parent=None, meta=None):
        """
        Given a :class:`pathlib.Path`, create a Node from that path as well as
        any children

        If the path is not a file or a dir, an :class:`AssertionError` is raised

        :param path:
            A :class:`pathlib.Path` that is either the ``content_path`` or a child of it.
        :param parent:
            Either another :class:`Node` or ``None``
        :param meta:
            A dict-like object that will be passed to a :class:`Config` instance
        """
        # path should be a pathlib object
        assert path.is_file() or path.is_dir()

        node = cls(path, parent=parent, meta=meta)

        if path.is_dir():
            for child in path.iterdir():
                ignored = False
                for glob in node.meta.get("ignore", []):
                    if child in path.glob(glob):
                        ignored = True
                        break
                if ignored:
                    continue

                if child.name in cls._meta_names and child.is_file():
                    with child.open() as co:
                        node.meta.load(co)
                else:
                    cls.from_path(child, node)

        return node

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.path_obj.name)

    @property
    def full_path(self):
        """
        Full path of node when deployed
        """
        if self.parent is None:
            return self.meta["deploy_path"]
        else:
            return str(pathlib.Path(self.parent.full_path, self.path_obj.name))

    @property
    def full_url(self):
        """Get full URL for node, including trailing slash"""
        if self.parent is None:
            base_url = self.meta.get("base_url", "/")
            if not base_url.startswith("/"):
                base_url = "/" + base_url
            if not base_url.endswith("/"):
                base_url = base_url + "/"

            return base_url
        elif self.is_leaf:
            if self.path_obj.name == self._index_file:
                name = ""
            elif self.path_obj.suffix in self._strip_exts:
                name = self.path_obj.stem
            else:
                name = self.path_obj.name

            return "".join([self.parent.full_url, name])
        else:
            return "".join([self.parent.full_url, self.path_obj.name, "/"])

    def walk(self, include_self=False):
        """Walk through Node tree"""
        if include_self:
            yield self

        for child in self.children.values():
            yield child
            for grandchild in child.walk():
                yield grandchild

    def render(self):
        """
        Process node and either create the directory or write contents of file to ``deploy_path``
        """
        if not self.is_leaf:
            pathlib.Path(self.full_path).mkdir(self._dir_mode)
            return

        file_obj = pathlib.Path(self.full_path)
        file_obj.touch(self._file_mode)

        content = self.get_content()

        with file_obj.open("w" if type(content) is str else "wb") as fo:
            fo.write(content)

    def process_meta(self):
        """
        Finds and processes the YAML fonrt matter at the top of a file

        If the file does not start with ``---\\n``, then it's assumed the file
        does not contain any meta YAML for us to process
        """
        if self._content_start is not None:
            # we've done this already
            return
        if not self.is_leaf:
            self._content_start = 0
            return

        found_header = False
        with self.path_obj.open("rb") as file_obj:
            while True:
                data = file_obj.read(100)
                try:
                    data = data.decode("utf-8")
                except UnicodeDecodeError:
                    self._content_start = 0
                    return

                if data == '':
                    # we've run out of file, either we didn't find a header or
                    # we're missing a footer
                    self._content_start = 0
                    return
                elif not found_header:
                    if data.startswith(self._meta_header):
                        found_header = True
                        found_meta = data[len(self._meta_header):]
                    else:
                        # if our token is not the first thing in the file, then
                        # it's not for us
                        self._content_start = 0
                        return
                else:
                    found_meta += data

                if self._meta_footer in found_meta:
                    idx = found_meta.index(self._meta_footer)
                    found_meta = found_meta[:idx]
                    break

        self._meta.load(found_meta)
        self._content_start = len(self._meta_header) + len(found_meta) + len(self._meta_footer)

    def get_content(self):
        """
        Get the actual content of the Node

        First calls :meth:`process_meta` to find the end any front matter that
        might be present and then returns the rest of the file

        If ``filter`` has been specified in :attr:`meta`, that filter will be
        used to further process the content.
        """
        if self._content is not None:
            return self._content

        self.process_meta()
        content_filter = self.meta.get("filter")
        with self.path_obj.open("rb") as file_obj:
            file_obj.seek(self._content_start)
            self._content = file_obj.read()
            try:
                self._content = self._content.decode("utf-8")
            except UnicodeDecodeError:
                return self._content

            if content_filter is not None:
                filter_module = import_module(content_filter)
                filter_glob = self.meta.get("filter-glob", filter_module.DEFAULT_GLOB)
                if self.path_obj in self.path_obj.parent.glob(filter_glob):
                    self._content = filter_module.content_filter(self, self._content)

        return self._content

    @property
    def meta(self):
        """
        Configuration object

        Automatically loads front-matter if applicable
        """
        self.process_meta()
        return self._meta

    @property
    def marks(self):
        """
        Marked sections from content

        Calls :meth:`get_content` to process content if that hasn't been done
        already
        """
        if self._marks is not None:
            return self._marks

        self._marks = {}
        # make sure that _marks gets populated
        self.get_content()

        return self._marks

    @property
    def data(self):
        """Extracts data from contents of file

        For example, a YAML file
        """
        if self.path_obj.is_dir():
            return

        if self._data is None:
            data = self.get_content()

            func = DATA_EXTRACTORS[self.path_obj.suffix]

            self._data = func(data)

        return self._data

    def add_child(self, child):
        """
        Add a child to the current Node

        If the child doesn't already have its :attr:`parent` set to this Node,
        then an :class:`AssertionError` is raised.
        """
        assert child.parent == self
        self.children[child.path_obj.name] = child

    @property
    def siblings(self):
        """Returns all children of the parent Node, except for itself"""
        return {k: v for k, v in self.parent.children.items() if v is not self}


def gen(settings):
    """
    Generate site

    Deletes ``deploy_path`` first.
    """
    shutil.rmtree(settings["deploy_path"], True)
    root_node = Node.from_path(pathlib.Path(settings["content_path"]), meta=settings)

    for item in root_node.walk(True):
        logger.info("Rendering %s", item.full_url)
        item.render()


def serve(settings):
    """
    Serves the generated site from ``deploy_path``

    Respects settings like ``base_url`` if present.
    """
    logger = logging.getLogger("exhibition.server")

    class ExhibitionHTTPRequestHandler(SimpleHTTPRequestHandler):
        def translate_path(self, path):
            path = path.strip("/")
            if settings.get("base_url"):
                base = settings["base_url"].strip("/")
                if not path.startswith(base):
                    return ""
                path = path.lstrip(base).strip("/")

            path = os.path.join(settings["deploy_path"], path)

            return path

    server_address = ('localhost', 8000)

    httpd = HTTPServer(server_address, ExhibitionHTTPRequestHandler)

    logger.warning("Listening on http://%s:%s", *server_address)
    t = threading.Thread(target=httpd.serve_forever, daemon=True)
    t.start()

    return (httpd, t)



from setuptools import setup, find_packages

with open("README.rst", "r") as f:
    long_desc = f.read()

setup(
    name="exhibition",
    version="0.0.3",
    author="Matt Molyneaux",
    author_email="moggers87+git@moggers87.co.uk",
    description="A Python static site generator",
    long_description=long_desc,
    url="https://github.com/moggers87/exhibition",
    download_url="https://pypi.org/project/exhibition/",
    packages=find_packages(),
    test_suite="exhibition.tests",
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Internet",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development",
    ],
    install_requires=[
        "click",
        "jinja2",
        "markdown",
        "ruamel.yaml",
        "typogrify",
    ],
    extras_require={
        "docs": [
            "sphinx",
            "sphinx_rtd_theme",
            "sphinx-click",
        ],
    },
    entry_points={
        "console_scripts": [
            "exhibit = exhibition.command:exhibition",
        ]
    },
)
